# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module provides the circuit drawing functionality used to display circuits visually.

.. currentmodule:: pennylane.drawer
.. autosummary::
    :toctree: api

"""

from .draw import draw, draw_mpl
from .tape_text import tape_text
from .tape_mpl import tape_mpl
from .mpldrawer import MPLDrawer
from .style import available_styles, use_style
